import random
import time
import threading
from typing import List, Dict, Any
from crypto_utils import raw_signature_to_der

from crypto_utils import (
    generate_identity_keys,
    load_peer_identity_key,
    sign_data,
    verify_signature,
    get_ephemeral_keypair,
    compute_ecdh_shared_key
)
from messages import (
    build_ping_request, parse_ping_request,
    build_ping_response, parse_ping_response,
    build_handshake_message, parse_handshake_message,
    compute_pfs_hash
)
import transmission

# ANSI colors
RED = "\033[91m"
GREEN = "\033[92m"
YELLOW = "\033[93m"
BLUE = "\033[94m"
RESET = "\033[0m"


class IcingProtocol:
    def __init__(self):
        # Identity keys
        self.identity_privkey, self.identity_pubkey = generate_identity_keys()

        # Peer identity
        self.peer_identity_pubkey_obj = None
        self.peer_identity_pubkey_bytes = None

        # Ephemeral keys
        self.ephemeral_privkey = None
        self.ephemeral_pubkey = None

        # Last computed shared secret (hex)
        self.shared_secret = None

        # For PFS: track session_number + last_shared_secret per peer identity
        # Key: bytes(64) peer identity pubkey
        # Value: (int session_number, str last_shared_secret_hex)
        self.pfs_history: Dict[bytes, (int, str)] = {}

        # Inbound messages are stored for manual or auto handling
        # Each entry: { 'type': str, 'raw': bytes, 'parsed': Any, 'connection': PeerConnection }
        self.inbound_messages: List[Dict[str, Any]] = []

        # Simple dictionary to track protocol flags
        self.state = {
            "ping_sent": False,
            "ping_received": False,
            "handshake_sent": False,
            "handshake_received": False,
        }

        # Auto-responder toggle
        self.auto_responder = False

        # Connections
        self.connections = []

        # Listening port
        self.local_port = random.randint(30000, 40000)
        self.server_listener = transmission.ServerListener(
            host="127.0.0.1",
            port=self.local_port,
            on_new_connection=self.on_new_connection,
            on_data_received=self.on_data_received
        )
        self.server_listener.start()

    # -------------------------------------------------------------------------
    # Transport callbacks
    # -------------------------------------------------------------------------

    def on_new_connection(self, conn: transmission.PeerConnection):
        print(f"{GREEN}[IcingProtocol]{RESET} New incoming connection.")
        self.connections.append(conn)

    def on_data_received(self, conn: transmission.PeerConnection, data: bytes):
        """
        Called whenever data arrives on any open PeerConnection.
        We'll parse and store the message, then handle automatically if auto_responder is on.
        """
        # Print data size in bits, not bytes
        bits_count = len(data)*8
        print(f"{GREEN}[RECV]{RESET} {bits_count} bits from peer: {data.hex()[:60]}{'...' if len(data.hex())>60 else ''}")

        # Attempt to parse Ping request
        if len(data) == 37:
            parsed = parse_ping_request(data)
            if parsed:
                nonce, version = parsed
                self.state["ping_received"] = True
                index = len(self.inbound_messages)
                msg = {
                    "type": "PING_REQUEST",
                    "raw": data,
                    "parsed": {"nonce": nonce, "version": version},
                    "connection": conn
                }
                self.inbound_messages.append(msg)
                print(f"{YELLOW}[NOTICE]{RESET} Stored inbound PING request (nonce={nonce.hex()}) at index={index}.")

                if self.auto_responder:
                    # Schedule an automatic response after 2 seconds
                    threading.Timer(2.0, self._auto_respond_ping, args=(index,)).start()

                return

        # Attempt to parse Ping response
        if len(data) == 9:
            parsed = parse_ping_response(data)
            if parsed:
                ts, version, answer_code = parsed
                index = len(self.inbound_messages)
                msg = {
                    "type": "PING_RESPONSE",
                    "raw": data,
                    "parsed": {"timestamp": ts, "version": version, "answer_code": answer_code},
                    "connection": conn
                }
                self.inbound_messages.append(msg)
                print(f"{YELLOW}[NOTICE]{RESET} Stored inbound PING response (answer_code={answer_code}) at index={index}.")
                return

        # Attempt to parse handshake
        if len(data) == 168:
            parsed = parse_handshake_message(data)
            if parsed:
                ts, ephemeral_pub, ephemeral_sig, pfs_hash = parsed
                self.state["handshake_received"] = True
                index = len(self.inbound_messages)
                msg = {
                    "type": "HANDSHAKE",
                    "raw": data,
                    "parsed": {
                        "ephemeral_pub": ephemeral_pub,
                        "ephemeral_sig": ephemeral_sig,
                        "timestamp": ts,
                        "pfs hash": pfs_hash
                    },
                    "connection": conn
                }
                self.inbound_messages.append(msg)
                print(f"{YELLOW}[NOTICE]{RESET} Stored inbound HANDSHAKE at index={index}. ephemeral_pub={ephemeral_pub.hex()[:20]}...")

                if self.auto_responder:
                    # Schedule an automatic handshake "response" after 2 seconds
                    threading.Timer(2.0, self._auto_respond_handshake, args=(index,)).start()

                return

        # Otherwise, unrecognized
        index = len(self.inbound_messages)
        msg = {
            "type": "UNKNOWN",
            "raw": data,
            "parsed": None,
            "connection": conn
        }
        self.inbound_messages.append(msg)
        print(f"{RED}[WARNING]{RESET} Unrecognized or malformed message stored at index={index}.")

    # -------------------------------------------------------------------------
    # Auto-responder helpers
    # -------------------------------------------------------------------------

    def _auto_respond_ping(self, index: int):
        """
        Called by a Timer to respond automatically to a PING_REQUEST after 2s.
        """
        print(f"{BLUE}[AUTO]{RESET} Delayed responding to PING at index={index}")
        self.respond_to_ping(index, answer_code=0)
        self.show_state()

    def _auto_respond_handshake(self, index: int):
        """
        Called by a Timer to handle inbound HANDSHAKE automatically.
        1) Generate ephemeral keys if not already set
        2) Compute ECDH with the inbound ephemeral pub (generate_ecdhe)
        3) Send our handshake back
        4) Show state
        """
        print(f"{BLUE}[AUTO]{RESET} Delayed ECDH process for HANDSHAKE at index={index}")

        # 1) Generate ephemeral keys if we don't have them
        if not self.ephemeral_privkey or not self.ephemeral_pubkey:
            self.generate_ephemeral_keys()

        # 2) Compute ECDH from inbound ephemeral pub
        self.generate_ecdhe(index)

        # 3) Send our handshake to the peer
        self.send_handshake()

        # 4) Show final state
        self.show_state()

    # -------------------------------------------------------------------------
    # Public Methods
    # -------------------------------------------------------------------------

    def connect_to_peer(self, port: int):
        conn = transmission.connect_to_peer("127.0.0.1", port, self.on_data_received)
        self.connections.append(conn)
        print(f"{GREEN}[IcingProtocol]{RESET} Outgoing connection to port {port} established.")

    def set_peer_identity(self, peer_pubkey_hex: str):
        pubkey_bytes = bytes.fromhex(peer_pubkey_hex)
        self.peer_identity_pubkey_obj = load_peer_identity_key(pubkey_bytes)
        self.peer_identity_pubkey_bytes = pubkey_bytes
        print(f"{GREEN}[IcingProtocol]{RESET} Stored peer identity pubkey (hex={peer_pubkey_hex[:16]}...).")

    def generate_ephemeral_keys(self):
        self.ephemeral_privkey, self.ephemeral_pubkey = get_ephemeral_keypair()
        print(f"{GREEN}[IcingProtocol]{RESET} Generated ephemeral key pair: pubkey={self.ephemeral_pubkey.hex()[:16]}...")

    def send_ping_request(self):
        if not self.connections:
            print(f"{RED}[ERROR]{RESET} No active connections.")
            return
        pkt = build_ping_request(version=0)
        self._send_packet(self.connections[0], pkt, "PING_REQUEST")
        self.state["ping_sent"] = True

    def send_handshake(self):
        """
        Build and send handshake:
          - 32-bit timestamp
          - ephemeral_pubkey (64 bytes, raw x||y)
          - ephemeral_signature (64 bytes, raw r||s)
          - pfs_hash (32 bytes)
          - 32-bit CRC
        """
        if not self.connections:
            print(f"{RED}[ERROR]{RESET} No active connections.")
            return
        if not self.ephemeral_privkey or not self.ephemeral_pubkey:
            print(f"{RED}[ERROR]{RESET} Ephemeral keys not generated.")
            return
        if self.peer_identity_pubkey_bytes is None:
            print(f"{RED}[ERROR]{RESET} Peer identity not set; needed for PFS tracking.")
            return

        # 1) Sign ephemeral_pubkey as r||s
        #    Instead of DER, we do raw r||s each 32 bytes
        sig_der = sign_data(self.identity_privkey, self.ephemeral_pubkey)
        # Convert DER -> (r, s) -> raw 64 bytes
        # Quick approach to parse DER using cryptography, or do a custom parse
        from cryptography.hazmat.primitives.asymmetric.utils import decode_dss_signature
        r_int, s_int = decode_dss_signature(sig_der)
        r_bytes = r_int.to_bytes(32, 'big')
        s_bytes = s_int.to_bytes(32, 'big')
        raw_signature = r_bytes + s_bytes  # 64 bytes

        # 2) PFS hash
        session_number, last_secret_hex = self.pfs_history.get(self.peer_identity_pubkey_bytes, (-1, ""))
        pfs = compute_pfs_hash(session_number, last_secret_hex)

        # 3) Build handshake
        timestamp_32 = int(time.time() * 1000) & 0xffffffff
        pkt = build_handshake_message(
            timestamp_32,
            self.ephemeral_pubkey,  # 64 bytes raw
            raw_signature,  # 64 bytes raw
            pfs  # 32 bytes
        )

        # 4) Send
        self._send_packet(self.connections[0], pkt, "HANDSHAKE")
        self.state["handshake_sent"] = True

    def enable_auto_responder(self, enable: bool):
        self.auto_responder = enable
        print(f"{BLUE}[AUTO]{RESET} Auto responder set to {enable}.")

    # -------------------------------------------------------------------------
    # Manual Responses
    # -------------------------------------------------------------------------

    def respond_to_ping(self, index: int, answer_code: int):
        """
        Manually respond to an inbound PING_REQUEST in inbound_messages[index].
        """
        if index < 0 or index >= len(self.inbound_messages):
            print(f"{RED}[ERROR]{RESET} Invalid index {index}.")
            return
        msg = self.inbound_messages[index]
        if msg["type"] != "PING_REQUEST":
            print(f"{RED}[ERROR]{RESET} inbound_messages[{index}] is not a PING_REQUEST.")
            return

        version = msg["parsed"]["version"]
        conn = msg["connection"]
        resp = build_ping_response(version, answer_code)
        self._send_packet(conn, resp, "PING_RESPONSE")
        print(f"{BLUE}[MANUAL]{RESET} Responded to ping with answer_code={answer_code}.")

    def generate_ecdhe(self, index: int):
        """
        Formerly 'respond_to_handshake'. Verifies ephemeral signature, computes ECDH,
        updates pfs_history, and stores shared_secret. Does NOT send a handshake back.
        """
        if index < 0 or index >= len(self.inbound_messages):
            print(f"{RED}[ERROR]{RESET} Invalid index {index}.")
            return
        msg = self.inbound_messages[index]
        if msg["type"] != "HANDSHAKE":
            print(f"{RED}[ERROR]{RESET} inbound_messages[{index}] is not a HANDSHAKE.")
            return

        # Parse fields
        # (timestamp, ephemeral_pub, ephemeral_sig, pfs_hash) = ...
        (ts32, ephemeral_pub, ephemeral_sig, pfs_val) = msg["parsed"]

        # 1) Verify ephemeral signature
        if not self.peer_identity_pubkey_obj:
            print(f"{RED}[ERROR]{RESET} Peer identity not set.")
            return
        # ephemeral_sig is raw r||s
        sig_der = raw_signature_to_der(ephemeral_sig)
        ok = verify_signature(self.peer_identity_pubkey_obj, sig_der, ephemeral_pub)
        if not ok:
            print(f"{RED}[ERROR]{RESET} Ephemeral signature invalid.")
            return
        print(f"{GREEN}[OK]{RESET} Ephemeral signature verified.")

        # 2) ECDH
        if not self.ephemeral_privkey:
            print(f"{YELLOW}[WARN]{RESET} No ephemeral_privkey. Cannot compute shared secret.")
            return
        shared = compute_ecdh_shared_key(self.ephemeral_privkey, ephemeral_pub)
        self.shared_secret = shared.hex()
        print(f"{GREEN}[OK]{RESET} Computed ECDH shared key = {self.shared_secret}")

        # 3) Update pfs_history
        # If we have an entry, increment session_number, store new secret
        # If none, create session_number=1, store new secret
        old_session, _ = self.pfs_history.get(self.peer_identity_pubkey_bytes, (-1, ""))
        new_session = 1 if old_session < 0 else old_session + 1
        self.pfs_history[self.peer_identity_pubkey_bytes] = (new_session, self.shared_secret)

    # -------------------------------------------------------------------------
    # Utility
    # -------------------------------------------------------------------------

    def _send_packet(self, conn: transmission.PeerConnection, data: bytes, label: str):
        bits_count = len(data) * 8
        print(f"{BLUE}[SEND]{RESET} {label} -> {bits_count} bits: {data.hex()[:60]}{'...' if len(data.hex())>60 else ''}")
        conn.send(data)

    def show_state(self):
        print(f"\n{YELLOW}=== Global State ==={RESET}")
        print(f"Listening Port: {self.local_port}")
        print(f"Identity PubKey: 512 bits => {self.identity_pubkey.hex()[:16]}...")

        if self.peer_identity_pubkey_bytes:
            print(f"Peer Identity PubKey: 512 bits => {self.peer_identity_pubkey_bytes.hex()[:16]}...")
        else:
            print("Peer Identity PubKey: [None]")

        print("\nEphemeral Keys:")
        if self.ephemeral_pubkey:
            print(f"  ephemeral_pubkey: 512 bits => {self.ephemeral_pubkey.hex()[:16]}...")
        else:
            print("  ephemeral_pubkey: [None]")

        print(f"\nShared Secret: {self.shared_secret if self.shared_secret else '[None]'}")

        print("\nProtocol Flags:")
        for k, v in self.state.items():
            print(f"  {k}: {v}")

        print("\nAuto Responder:", self.auto_responder)

        print("\nActive Connections:")
        for i, c in enumerate(self.connections):
            print(f"  [{i}] Alive={c.alive}")

        print("\nInbound Message Queue:")
        for i, m in enumerate(self.inbound_messages):
            print(f"  [{i}] type={m['type']} len={len(m['raw'])} bytes => {len(m['raw']) * 8} bits")
        print()

    def stop(self):
        self.server_listener.stop()
        for c in self.connections:
            c.close()
        self.connections.clear()
        self.inbound_messages.clear()
        print(f"{RED}[STOP]{RESET} Protocol stopped.")
